"""
Data Types for Sutton SignWriting core functionality.
"""

from typing import (
    Dict,
    TypedDict,
    List,
    Union,
    NotRequired,
)

# Query-related types


class QuerySignboxSymbol(TypedDict):
    """A symbol in a signbox query."""

    symbol: str
    """A symbol identifier."""
    coord: NotRequired[List[int]]
    """An optional coordinate [x, y]."""


class QuerySignboxRange(TypedDict):
    """A symbol range in a signbox query."""

    range: List[str]
    """An array of two symbols defining the range [start, end]."""
    coord: NotRequired[List[int]]
    """An optional coordinate [x, y]."""


class QuerySignboxOr(TypedDict):
    """An OR group of symbols or ranges in a signbox query."""

    or_list: List[Union[str, List[str]]]
    """An array of symbol strings and/or range arrays."""
    coord: NotRequired[List[int]]
    """An optional coordinate [x, y]."""


QuerySignboxElement = Union[QuerySignboxSymbol, QuerySignboxRange, QuerySignboxOr]
"""A signbox query element as a symbol, range, or an OR group of symbols and ranges."""


QueryPrefixSymbol = str
"""A symbol identifier in a prefix query."""

QueryPrefixRange = List[str]
"""A symbol range in a prefix query."""

QueryPrefixOr = List[Union[str, List[str]]]
"""An OR group of symbols or ranges in a prefix query."""

QueryPrefixElement = Union[QueryPrefixSymbol, QueryPrefixRange, QueryPrefixOr]
"""A prefix query element as a :py:data:`.QueryPrefixSymbol`, :py:data:`.QueryPrefixRange`, or :py:data:`.QueryPrefixOr`."""


class QueryPrefix(TypedDict):
    """Prefix elements for a query."""

    required: bool
    """Whether the prefix is required."""
    parts: NotRequired[List[QueryPrefixElement]]
    """A list of symbols, ranges, or OR groups."""


class QueryObject(TypedDict):
    """
    Object of query elements with regular expression identification.
    """

    query: bool
    """Required true for query object."""
    prefix: NotRequired[QueryPrefix]
    """An object for prefix elements."""
    signbox: NotRequired[List[QuerySignboxElement]]
    """Array of objects for symbols, ranges, and list of symbols or ranges, with optional coordinates."""
    variance: NotRequired[int]
    """Amount that x or y coordinates can vary and find a match, defaults to 20."""
    style: NotRequired[bool]
    """Boolean value for including style string in matches."""


# Column and layout types


class ColumnPunctuation(TypedDict):
    """Punctuation options for columns."""

    spacing: NotRequired[bool]
    """Whether to add spacing for punctuation."""
    pad: NotRequired[int]
    """Padding for punctuation."""
    pull: NotRequired[bool]
    """Whether to pull punctuation closer."""


class DetailSym(TypedDict):
    """
    A symbol-specific style configuration for custom colors.
    """

    index: int
    """The index of the symbol."""
    detail: List[str]
    """Array of CSS names or hex colors for line and optional fill."""


class StyleObject(TypedDict):
    """
    The elements of a style string.
    """

    colorize: NotRequired[bool]
    """Boolean to use standardized colors for symbol groups."""
    padding: NotRequired[int]
    """Integer value for padding around symbol or sign."""
    background: NotRequired[str]
    """CSS name or hex color for background."""
    detail: NotRequired[List[str]]
    """Array for CSS name or hex color for line and optional fill."""
    zoom: NotRequired[Union[int, float, str]]
    """Decimal value for zoom level."""
    detailsym: NotRequired[List[DetailSym]]
    """Custom colors for individual symbols: list of {index: int, detail: List[str]}."""
    classes: NotRequired[str]
    """List of class names separated with spaces used for SVG."""
    id: NotRequired[str]
    """ID name used for SVG."""


class ColumnOptions(TypedDict):
    """
    Options for column layout.
    """

    height: NotRequired[int]
    """The height of the columns."""
    width: NotRequired[int]
    """The widths of the columns."""
    offset: NotRequired[int]
    """The lane offset for left and right lanes."""
    pad: NotRequired[int]
    """Amount of padding before and after signs as well as at top, left, and right of columns."""
    margin: NotRequired[int]
    """Amount of space at bottom of column that is not available."""
    dynamic: NotRequired[bool]
    """Enables variable width columns."""
    background: NotRequired[str]
    """Background color for columns."""
    style: NotRequired[StyleObject]
    """An object of style options."""
    punctuation: NotRequired[ColumnPunctuation]
    """An object of punctuation options."""


class SegmentInfo(TypedDict):
    """
    Information about a text segment.
    """

    minX: int
    """The min x value within the segment."""
    minY: int
    """The min y value within the segment."""
    width: int
    """The width of the text segment."""
    height: int
    """The height of the text segment."""
    lane: int
    """Left as -1, Middle as 0, Right as 1."""
    padding: int
    """The padding of the text segment affects colored background."""
    segment: str
    """'sign' or 'symbol'."""
    zoom: Union[int, float, str]
    """The zoom size of the segment."""


class ColumnSegment(SegmentInfo):
    """A segment in a column layout (extends SegmentInfo with position and text)."""

    x: int
    """The x position in the column."""
    y: int
    """The y position in the column."""
    text: str
    """The text of the sign or symbol with optional style string."""


ColumnData = List[ColumnSegment]


class ColumnsResult(TypedDict):
    """
    The result of the columns functions, containing column layout data.
    """

    options: NotRequired[ColumnOptions]
    """Column layout options."""
    widths: NotRequired[List[int]]
    """Widths of each column."""
    columns: NotRequired[List[List[ColumnSegment]]]
    """Array of columns, each containing an array of segments."""


class SymbolObject(TypedDict):
    """
    The elements of a symbol string.
    """

    symbol: NotRequired[str]
    """Symbol identifier."""
    coord: NotRequired[List[int]]
    """X, y coordinate."""
    style: NotRequired[str]
    """Style string."""


class SignSpatial(TypedDict):
    """A spatial symbol in a sign."""

    symbol: str
    """Symbol identifier."""
    coord: List[int]
    """Coordinate [x, y]."""


class SignObject(TypedDict):
    """
    The elements of a sign string.
    """

    sequence: NotRequired[List[str]]
    """Array of symbols."""
    box: NotRequired[str]
    """Signbox marker or lane."""
    max: NotRequired[List[int]]
    """Preprocessed x, y coordinate."""
    spatials: NotRequired[List[SignSpatial]]
    """Array of symbols with coordinates."""
    style: NotRequired[str]
    """Style string."""


class SpecialToken(TypedDict):
    index: int
    name: str
    value: str


class TokenizerMappings(TypedDict):
    i2s: Dict[int, str]
    s2i: Dict[str, int]


__all__ = [
    "ColumnData",
    "ColumnOptions",
    "ColumnPunctuation",
    "ColumnSegment",
    "ColumnsResult",
    "QueryObject",
    "QueryPrefixSymbol",
    "QueryPrefixRange",
    "QueryPrefixOr",
    "QueryPrefixElement",
    "QueryPrefix",
    "QuerySignboxElement",
    "QuerySignboxOr",
    "QuerySignboxRange",
    "QuerySignboxSymbol",
    "SegmentInfo",
    "SignObject",
    "SignSpatial",
    "DetailSym",
    "StyleObject",
    "SymbolObject",
    "SpecialToken",
    "TokenizerMappings",
]



"""
sutton_signwriting_core – public entry point.
"""

from __future__ import annotations

from .symid_arr import symid_arr

from .datatypes import (
    ColumnData,
    ColumnOptions,
    ColumnPunctuation,
    ColumnSegment,
    QueryObject,
    QueryPrefixSymbol,
    QueryPrefixRange,
    QueryPrefixOr,
    QueryPrefixElement,
    QueryPrefix,
    QuerySignboxElement,
    QuerySignboxOr,
    QuerySignboxRange,
    QuerySignboxSymbol,
    SegmentInfo,
    SignObject,
    SignSpatial,
    DetailSym,
    StyleObject,
    SymbolObject,
    SpecialToken,
    TokenizerMappings,
)

from .regex import (
    fsw_pattern_null,
    fsw_pattern_symbol,
    fsw_pattern_number,
    fsw_pattern_sort,
    fsw_pattern_box,
    fsw_pattern_null_or_symbol,
    fsw_pattern_prefix,
    fsw_pattern_coord,
    fsw_pattern_spatial,
    fsw_pattern_signbox,
    fsw_pattern_sign,
    fsw_pattern_sortable,
    fswquery_pattern_null,
    fswquery_pattern_base,
    fswquery_pattern_coord,
    fswquery_pattern_var,
    fswquery_pattern_symbol,
    fswquery_pattern_null_or_symbol,
    fswquery_pattern_range,
    fswquery_pattern_item,
    fswquery_pattern_list,
    fswquery_pattern_prefix,
    fswquery_pattern_signbox,
    fswquery_pattern_full,
    style_pattern_colorize,
    style_pattern_colorhex,
    style_pattern_colorname,
    style_pattern_padding,
    style_pattern_zoom,
    style_pattern_classbase,
    style_pattern_id,
    style_pattern_colorbase,
    style_pattern_color,
    style_pattern_colors,
    style_pattern_background,
    style_pattern_detail,
    style_pattern_detailsym,
    style_pattern_classes,
    style_pattern_full_groups,
    style_pattern_full,
    swu_pattern_null,
    swu_pattern_symbol,
    swu_pattern_number,
    swu_pattern_sort,
    swu_pattern_box,
    swu_pattern_null_or_symbol,
    swu_pattern_prefix,
    swu_pattern_coord,
    swu_pattern_spatial,
    swu_pattern_signbox,
    swu_pattern_sign,
    swu_pattern_sortable,
    swuquery_pattern_null,
    swuquery_pattern_base,
    swuquery_pattern_coord,
    swuquery_pattern_var,
    swuquery_pattern_symbol,
    swuquery_pattern_null_or_symbol,
    swuquery_pattern_range,
    swuquery_pattern_item,
    swuquery_pattern_list,
    swuquery_pattern_prefix,
    swuquery_pattern_signbox,
    swuquery_pattern_full,
)

from .convert import (
    swu_to_mark,
    mark_to_swu,
    swu_to_num,
    num_to_swu,
    swu_to_coord,
    coord_to_swu,
    fsw_to_coord,
    coord_to_fsw,
    swu_to_code,
    code_to_swu,
    swu_to_id,
    id_to_swu,
    key_to_id,
    id_to_key,
    swu_to_key,
    key_to_swu,
    swu_to_fsw,
    fsw_to_swu,
    symid_max,
    symid_min,
    symid_to_key,
    key_to_symid,
)

from .fsw import (
    fsw_structure_kind,
    fsw_structure_category,
    fsw_structure_group,
    fsw_structure_ranges,
    fsw_is_type,
    fsw_colors,
    fsw_colorize,
    fsw_parse_symbol,
    fsw_parse_sign,
    fsw_parse_text,
    fsw_compose_symbol,
    fsw_compose_sign,
    fsw_info,
    fsw_column_defaults,
    fsw_column_defaults_merge,
    fsw_columns,
    fsw_tokenize,
    fsw_detokenize,
    fsw_chunk_tokens,
    FSWTokenizer,
)

from .fswquery import (
    fswquery_parse,
    fswquery_compose,
    fsw_to_query,
    fswquery_range,
    fswquery_regex,
    fswquery_results,
    fswquery_lines,
)

from .style import (
    style_parse,
    style_compose,
    style_rgb_to_hex,
    style_rgba_to_hex,
    style_merge,
)

from .swu import (
    swu_structure_kind,
    swu_structure_category,
    swu_structure_group,
    swu_structure_ranges,
    swu_is_type,
    swu_colors,
    swu_colorize,
    swu_parse_symbol,
    swu_parse_sign,
    swu_parse_text,
    swu_compose_symbol,
    swu_compose_sign,
    swu_info,
    swu_column_defaults,
    swu_column_defaults_merge,
    swu_columns,
)

from .swuquery import (
    swuquery_parse,
    swuquery_compose,
    swu_to_query,
    swuquery_range,
    swuquery_regex,
    swuquery_results,
    swuquery_lines,
)


__all__ = [
    # data types
    "ColumnData",
    "ColumnOptions",
    "ColumnPunctuation",
    "ColumnSegment",
    "QueryObject",
    "QueryPrefixSymbol",
    "QueryPrefixRange",
    "QueryPrefixOr",
    "QueryPrefixElement",
    "QueryPrefix",
    "QuerySignboxElement",
    "QuerySignboxOr",
    "QuerySignboxRange",
    "QuerySignboxSymbol",
    "SegmentInfo",
    "SignObject",
    "SignSpatial",
    "DetailSym",
    "StyleObject",
    "SymbolObject",
    "SpecialToken",
    "TokenizerMappings",
    # FSW patterns
    "fsw_pattern_null",
    "fsw_pattern_symbol",
    "fsw_pattern_number",
    "fsw_pattern_sort",
    "fsw_pattern_box",
    "fsw_pattern_null_or_symbol",
    "fsw_pattern_prefix",
    "fsw_pattern_coord",
    "fsw_pattern_spatial",
    "fsw_pattern_signbox",
    "fsw_pattern_sign",
    "fsw_pattern_sortable",
    # FSW query patterns
    "fswquery_pattern_null",
    "fswquery_pattern_base",
    "fswquery_pattern_coord",
    "fswquery_pattern_var",
    "fswquery_pattern_symbol",
    "fswquery_pattern_null_or_symbol",
    "fswquery_pattern_range",
    "fswquery_pattern_item",
    "fswquery_pattern_list",
    "fswquery_pattern_prefix",
    "fswquery_pattern_signbox",
    "fswquery_pattern_full",
    # Style patterns
    "style_pattern_colorize",
    "style_pattern_colorhex",
    "style_pattern_colorname",
    "style_pattern_padding",
    "style_pattern_zoom",
    "style_pattern_classbase",
    "style_pattern_id",
    "style_pattern_colorbase",
    "style_pattern_color",
    "style_pattern_colors",
    "style_pattern_background",
    "style_pattern_detail",
    "style_pattern_detailsym",
    "style_pattern_classes",
    "style_pattern_full_groups",
    "style_pattern_full",
    # SWU patterns
    "swu_pattern_null",
    "swu_pattern_symbol",
    "swu_pattern_number",
    "swu_pattern_sort",
    "swu_pattern_box",
    "swu_pattern_null_or_symbol",
    "swu_pattern_prefix",
    "swu_pattern_coord",
    "swu_pattern_spatial",
    "swu_pattern_signbox",
    "swu_pattern_sign",
    "swu_pattern_sortable",
    # SWU query patterns
    "swuquery_pattern_null",
    "swuquery_pattern_base",
    "swuquery_pattern_coord",
    "swuquery_pattern_var",
    "swuquery_pattern_symbol",
    "swuquery_pattern_null_or_symbol",
    "swuquery_pattern_range",
    "swuquery_pattern_item",
    "swuquery_pattern_list",
    "swuquery_pattern_prefix",
    "swuquery_pattern_signbox",
    "swuquery_pattern_full",
    # Text conversion
    "swu_to_fsw",
    "fsw_to_swu",
    # Symbol ↔ key ↔ id
    "key_to_swu",
    "swu_to_key",
    "key_to_id",
    "id_to_key",
    # Markers / numbers / coordinates
    "swu_to_mark",
    "mark_to_swu",
    "swu_to_num",
    "num_to_swu",
    "swu_to_coord",
    "coord_to_swu",
    "fsw_to_coord",
    "coord_to_fsw",
    # Low-level symbol helpers
    "swu_to_code",
    "code_to_swu",
    "swu_to_id",
    "id_to_swu",
    # Symid helpers
    "symid_max",
    "symid_min",
    "symid_to_key",
    "key_to_symid",
    # Constant
    "symid_arr",
    # FSW
    "fsw_structure_kind",
    "fsw_structure_category",
    "fsw_structure_group",
    "fsw_structure_ranges",
    "fsw_is_type",
    "fsw_colors",
    "fsw_colorize",
    "fsw_parse_symbol",
    "fsw_parse_sign",
    "fsw_parse_text",
    "fsw_compose_symbol",
    "fsw_compose_sign",
    "fsw_info",
    "fsw_column_defaults",
    "fsw_column_defaults_merge",
    "fsw_columns",
    "fsw_tokenize",
    "fsw_detokenize",
    "fsw_chunk_tokens",
    "fsw_create_tokenizer",
    "FSWTokenizer",
    # FSW Query
    "fswquery_parse",
    "fswquery_compose",
    "fsw_to_query",
    "fswquery_range",
    "fswquery_regex",
    "fswquery_results",
    "fswquery_lines",
    # Style
    "style_parse",
    "style_compose",
    "style_rgb_to_hex",
    "style_rgba_to_hex",
    "style_merge",
    # SWU
    "swu_structure_kind",
    "swu_structure_category",
    "swu_structure_group",
    "swu_structure_ranges",
    "swu_is_type",
    "swu_colors",
    "swu_colorize",
    "swu_parse_symbol",
    "swu_parse_sign",
    "swu_parse_text",
    "swu_compose_symbol",
    "swu_compose_sign",
    "swu_info",
    "swu_column_defaults",
    "swu_column_defaults_merge",
    "swu_columns",
    # SWU Query
    "swuquery_parse",
    "swuquery_compose",
    "swu_to_query",
    "swuquery_range",
    "swuquery_regex",
    "swuquery_results",
    "swuquery_lines",
]

__version__ = "1.1.0"



import re

# ----------------------------
# Utility
# ----------------------------


def _non_capturing(pattern: str) -> str:
    """Convert a regex with capturing groups into a non-capturing equivalent."""
    return re.sub(r"\((?!\?)", "(?:", pattern)


# ----------------------------
# FSW patterns
# ----------------------------

fsw_pattern_null = r"S00000"
""""""
fsw_pattern_symbol = r"S[123][0-9a-f]{2}[0-5][0-9a-f]"
""""""
fsw_pattern_number = r"[0-9]{3}"
""""""
fsw_pattern_sort = r"A"
""""""
fsw_pattern_box = r"[BLMR]"
""""""

fsw_pattern_null_or_symbol = rf"(?:{fsw_pattern_null}|{fsw_pattern_symbol})"
""""""
fsw_pattern_prefix = rf"(?:{fsw_pattern_sort}{fsw_pattern_null_or_symbol}+)"
""""""
fsw_pattern_coord = rf"{fsw_pattern_number}x{fsw_pattern_number}"
""""""
fsw_pattern_spatial = rf"{fsw_pattern_symbol}{fsw_pattern_coord}"
""""""
fsw_pattern_signbox = rf"{fsw_pattern_box}{fsw_pattern_coord}(?:{fsw_pattern_spatial})*"
""""""
fsw_pattern_sign = rf"{fsw_pattern_prefix}?{fsw_pattern_signbox}"
""""""
fsw_pattern_sortable = rf"{fsw_pattern_prefix}{fsw_pattern_signbox}"
""""""


# ----------------------------
# FSW Query patterns
# ----------------------------

fswquery_pattern_null = fsw_pattern_null
""""""
fswquery_pattern_base = r"[123][0-9a-f]{2}"
""""""
fswquery_pattern_coord = rf"(?:{fsw_pattern_coord})?"
""""""
fswquery_pattern_var = r"V[0-9]+"
""""""

fswquery_pattern_symbol = rf"S{fswquery_pattern_base}[0-5u][0-9a-fu]"
""""""
fswquery_pattern_null_or_symbol = (
    rf"(?:{fswquery_pattern_null}|{fswquery_pattern_symbol})"
)
""""""
fswquery_pattern_range = rf"R{fswquery_pattern_base}t{fswquery_pattern_base}"
""""""
fswquery_pattern_item = (
    rf"(?:{fswquery_pattern_null}|{fswquery_pattern_symbol}|{fswquery_pattern_range})"
)
""""""
fswquery_pattern_list = rf"{fswquery_pattern_item}(?:o{fswquery_pattern_item})*"
""""""
fswquery_pattern_prefix = rf"(?:A(?:{fswquery_pattern_list})+)?T"
""""""
fswquery_pattern_signbox = rf"(?:{fswquery_pattern_list}{fswquery_pattern_coord})*"
""""""
fswquery_pattern_full = rf"Q({fswquery_pattern_prefix})?({fswquery_pattern_signbox})?({fswquery_pattern_var})?(-?)"
""""""

# ----------------------------
# Style patterns
# ----------------------------
style_pattern_colorize = r"C"
""""""
style_pattern_colorhex = r"(?:[0-9a-fA-F]{3}){1,2}"
""""""
style_pattern_colorname = r"[a-zA-Z]+"
""""""
style_pattern_padding = r"P[0-9]{2}"
""""""
style_pattern_zoom = r"Z(?:[0-9]+(?:\.[0-9]+)?|x)"
""""""
style_pattern_classbase = r"-?[_a-zA-Z][_a-zA-Z0-9-]{0,100}"
""""""
style_pattern_id = r"[a-zA-Z][_a-zA-Z0-9-]{0,100}"
""""""

style_pattern_colorbase = rf"(?:{style_pattern_colorhex}|{style_pattern_colorname})"
""""""
style_pattern_color = rf"_{style_pattern_colorbase}_"
""""""
style_pattern_colors = rf"_{style_pattern_colorbase}(?:,{style_pattern_colorbase})?_"
""""""
style_pattern_background = rf"G{style_pattern_color}"
""""""
style_pattern_detail = rf"D{style_pattern_colors}"
""""""
style_pattern_detailsym = rf"D[0-9]{{2}}{style_pattern_colors}"
""""""
style_pattern_classes = rf"{style_pattern_classbase}(?: {style_pattern_classbase})*"
""""""
style_pattern_full_groups = (
    rf"-({style_pattern_colorize})?({style_pattern_padding})?({style_pattern_background})?"
    rf"({style_pattern_detail})?({style_pattern_zoom})?(?:-((?:{style_pattern_detailsym})*))?"
    rf"(?:-({style_pattern_classes})?!(?:({style_pattern_id})!)?)?"
)
""""""
style_pattern_full = _non_capturing(style_pattern_full_groups)
""""""

# ----------------------------
# SWU patterns
# ----------------------------

swu_pattern_null = "\\U00040000"
""""""
swu_pattern_symbol = "[\\U00040001-\\U0004F480]"
""""""
swu_pattern_number = "[\\U0001D80C-\\U0001D9FF]"
""""""
swu_pattern_sort = "\\U0001D800"
""""""
swu_pattern_box = "[\\U0001D801-\\U0001D804]"
""""""

swu_pattern_null_or_symbol = rf"(?:{swu_pattern_null}|{swu_pattern_symbol})"
""""""
swu_pattern_prefix = rf"(?:{swu_pattern_sort}{swu_pattern_null_or_symbol}+)"
""""""
swu_pattern_coord = rf"{swu_pattern_number}{{2}}"
""""""
swu_pattern_spatial = rf"{swu_pattern_symbol}{swu_pattern_coord}"
""""""
swu_pattern_signbox = rf"{swu_pattern_box}{swu_pattern_coord}(?:{swu_pattern_spatial})*"
""""""
swu_pattern_sign = rf"{swu_pattern_prefix}?{swu_pattern_signbox}"
""""""
swu_pattern_sortable = rf"{swu_pattern_prefix}{swu_pattern_signbox}"
""""""


# ----------------------------
# SWU Query patterns
# ----------------------------
swuquery_pattern_null = swu_pattern_null
""""""
swuquery_pattern_base = swu_pattern_symbol
""""""
swuquery_pattern_coord = rf"(?:{swu_pattern_number}{{2}})?"
""""""
swuquery_pattern_var = r"V[0-9]+"
""""""

swuquery_pattern_symbol = rf"{swuquery_pattern_base}f?r?"
""""""
swuquery_pattern_null_or_symbol = (
    rf"(?:{swuquery_pattern_null}|{swuquery_pattern_symbol})"
)
""""""
swuquery_pattern_range = rf"R{swuquery_pattern_base}{swuquery_pattern_base}"
""""""
swuquery_pattern_item = (
    rf"(?:{swuquery_pattern_null}|{swuquery_pattern_symbol}|{swuquery_pattern_range})"
)
""""""
swuquery_pattern_list = rf"{swuquery_pattern_item}(?:o{swuquery_pattern_item})*"
""""""
swuquery_pattern_prefix = rf"(?:A(?:{swuquery_pattern_list})+)?T"
""""""
swuquery_pattern_signbox = rf"(?:{swuquery_pattern_list}{swuquery_pattern_coord})*"
""""""
swuquery_pattern_full = rf"Q({swuquery_pattern_prefix})?({swuquery_pattern_signbox})?({swuquery_pattern_var})?(-?)"
""""""

# ----------------------------
# Exports
# ----------------------------
__all__ = [
    # FSW patterns
    "fsw_pattern_null",
    "fsw_pattern_symbol",
    "fsw_pattern_number",
    "fsw_pattern_sort",
    "fsw_pattern_box",
    "fsw_pattern_null_or_symbol",
    "fsw_pattern_prefix",
    "fsw_pattern_coord",
    "fsw_pattern_spatial",
    "fsw_pattern_signbox",
    "fsw_pattern_sign",
    "fsw_pattern_sortable",
    # FSW Query patterns
    "fswquery_pattern_null",
    "fswquery_pattern_base",
    "fswquery_pattern_coord",
    "fswquery_pattern_var",
    "fswquery_pattern_symbol",
    "fswquery_pattern_null_or_symbol",
    "fswquery_pattern_range",
    "fswquery_pattern_item",
    "fswquery_pattern_list",
    "fswquery_pattern_prefix",
    "fswquery_pattern_signbox",
    "fswquery_pattern_full",
    # Style patterns
    "style_pattern_colorize",
    "style_pattern_colorhex",
    "style_pattern_colorname",
    "style_pattern_padding",
    "style_pattern_zoom",
    "style_pattern_classbase",
    "style_pattern_id",
    "style_pattern_colorbase",
    "style_pattern_color",
    "style_pattern_colors",
    "style_pattern_background",
    "style_pattern_detail",
    "style_pattern_detailsym",
    "style_pattern_classes",
    "style_pattern_full_groups",
    "style_pattern_full",
    # SWU patterns
    "swu_pattern_null",
    "swu_pattern_symbol",
    "swu_pattern_number",
    "swu_pattern_sort",
    "swu_pattern_box",
    "swu_pattern_null_or_symbol",
    "swu_pattern_prefix",
    "swu_pattern_coord",
    "swu_pattern_spatial",
    "swu_pattern_signbox",
    "swu_pattern_sign",
    "swu_pattern_sortable",
    # SWU Query patterns
    "swuquery_pattern_null",
    "swuquery_pattern_base",
    "swuquery_pattern_coord",
    "swuquery_pattern_var",
    "swuquery_pattern_symbol",
    "swuquery_pattern_null_or_symbol",
    "swuquery_pattern_range",
    "swuquery_pattern_item",
    "swuquery_pattern_list",
    "swuquery_pattern_prefix",
    "swuquery_pattern_signbox",
    "swuquery_pattern_full",
]



"""
The swuquery module contains functions for handling the SWU query language.

Query Language definition: https://datatracker.ietf.org/doc/id/draft-slevinski-formal-signwriting-10.html#name-query-language
"""

import re
from typing import List, Optional, Union, cast

from .datatypes import (
    QueryObject,
    QueryPrefixElement,
    QueryPrefix,
    QuerySignboxElement,
    QuerySignboxOr,
    QuerySignboxRange,
    QuerySignboxSymbol,
)

from .convert import (
    swu_to_coord,
    coord_to_swu,
    num_to_swu,
    swu_to_key,
    key_to_swu,
)

from .regex import (
    style_pattern_full,
    swu_pattern_sort,
    swu_pattern_prefix,
    swu_pattern_box,
    swu_pattern_coord,
    swu_pattern_symbol,
    swu_pattern_null_or_symbol,
    swu_pattern_signbox,
    swuquery_pattern_coord,
    swuquery_pattern_var,
    swuquery_pattern_list,
    swuquery_pattern_item,
    swuquery_pattern_range,
    swuquery_pattern_symbol,
    swuquery_pattern_full,
)

from .swu import swu_parse_sign


# ----------------------------
# SWU Query Parsing
# ----------------------------


def _swuquery_parse_prefix(text: str) -> QueryPrefix:
    if text == "T":
        return {"required": True}
    parts = re.findall(swuquery_pattern_list, text)
    processed_parts: List[QueryPrefixElement] = []
    for part in parts:
        if "o" in part:
            or_parts = re.findall(swuquery_pattern_item, part)
            processed: list[Union[str, list[str]]] = ["or_list"]
            for or_part in or_parts:
                if or_part[0] != "R":
                    processed.append(or_part)
                else:
                    processed.append([or_part[1], or_part[2]])
            processed_parts.append(processed)
        else:
            if part[0] != "R":
                processed_parts.append(part)
            else:
                processed_parts.append([part[1], part[2]])
    return {"required": True, "parts": processed_parts}


def _swuquery_parse_signbox(text: str) -> List[QuerySignboxElement]:
    """
    Parse a signbox query string into a list of query elements.

    Args:
        text: A string representing the signbox query.

    Returns:
        A list of QuerySignboxElement (symbols, ranges, or OR groups).
    """
    items = re.findall(rf"{swuquery_pattern_list}{swuquery_pattern_coord}", text)
    processed: List[QuerySignboxElement] = []

    for item in items:
        coord_match = re.search(rf"{swuquery_pattern_coord}$", item)
        coord_str = coord_match.group(0) if coord_match else ""
        front = item[: -len(coord_str)] if coord_str else item
        coord = swu_to_coord(coord_str) if coord_str else None

        if "o" in front:
            or_parts = front.split("o")
            or_list = []
            for or_part in or_parts:
                if "R" not in or_part:
                    or_list.append(or_part)
                else:
                    or_list.append([or_part[1], or_part[2]])
            query_or: QuerySignboxOr = {"or_list": or_list}
            if coord is not None:
                query_or["coord"] = coord
            processed.append(query_or)
        elif "R" not in front:
            query_symbol: QuerySignboxSymbol = {"symbol": front}
            if coord is not None:
                query_symbol["coord"] = coord
            processed.append(query_symbol)
        else:
            query_range: QuerySignboxRange = {"range": [front[1], front[2]]}
            if coord is not None:
                query_range["coord"] = coord
            processed.append(query_range)

    return processed


def swuquery_parse(swu_query_string: str) -> QueryObject:
    """
    Parse an SWU query string to a structured dictionary.

    Args:
        swu_query_string: an SWU query string

    Returns:
        Dictionary representing the query structure

    Example:
        >>> swuquery_parse('QA񀀁R񀀁񆆑񆇡T񆀁R񀀁񀇱𝤆𝤆V5-')
        {'query': True,
         'prefix': {'required': True,
                    'parts': ['񀀁', ['񀀁', '񆆑'], '񆇡']},
         'signbox': [{'symbol': '񆀁'},
                     {'range': ['񀀁', '񀇱'], 'coord': [500, 500]}],
         'variance': 5,
         'style': True}
    """
    pattern = re.compile(rf"^{swuquery_pattern_full}$")
    m = pattern.match(swu_query_string)
    if not m:
        return {"query": False}

    result: QueryObject = {"query": True}
    prefix = _swuquery_parse_prefix(m.group(1)) if m.group(1) else None
    if prefix is not None:
        result["prefix"] = prefix
    signbox = _swuquery_parse_signbox(m.group(2)) if m.group(2) else None
    if signbox is not None:
        result["signbox"] = signbox
    variance = int(m.group(3)[1:]) if m.group(3) else None
    if variance is not None:
        result["variance"] = variance
    style = True if m.group(4) else None
    if style is not None:
        result["style"] = style

    return result


# ----------------------------
# SWU Query Compose
# ----------------------------


def swuquery_compose(swu_query_object: QueryObject) -> Optional[str]:
    """
    Function to compose SWU query string from object.

    Args:
        swu_query_object: Dictionary of type QueryObject

    Returns:
        SWU query string

    Example:
        >>> swuquery_compose({
        ...     'query': True,
        ...     'prefix': {
        ...         'required': True,
        ...         'parts': [
        ...             '񀀁',
        ...             ['񀀁', '񆆑'],
        ...             '񆇡'
        ...         ]
        ...     },
        ...     'signbox': [
        ...         {'symbol': '񆀁'},
        ...         {'range': ['񀀁', '񀇱'], 'coord': [500, 500]}
        ...     ],
        ...     'variance': 5,
        ...     'style': True
        ... })
        'QA񀀁R񀀁񆆑񆇡T񆀁R񀀁񀇱𝤆𝤆V5-'
    """
    if not swu_query_object or not swu_query_object.get("query"):
        return None

    query = "Q"

    prefix_obj = swu_query_object.get("prefix")
    if prefix_obj and prefix_obj.get("required"):
        parts = prefix_obj.get("parts")
        if isinstance(parts, list):
            query += "A"
            for part in parts:
                if isinstance(part, str):
                    query += part
                elif isinstance(part, list) and len(part) == 2:
                    query += f"R{part[0]}{part[1]}"
                elif isinstance(part, list) and len(part) > 2 and part[0] == "or_list":
                    or_parts = part[1:]
                    or_strs = []
                    for or_part in or_parts:
                        if isinstance(or_part, str):
                            or_strs.append(or_part)
                        elif isinstance(or_part, list) and len(or_part) == 2:
                            or_strs.append(f"R{or_part[0]}{or_part[1]}")
                    query += "o".join(or_strs)
        query += "T"

    signbox = swu_query_object.get("signbox")
    if isinstance(signbox, list):
        for part_sb in signbox:
            out = ""
            if isinstance(part_sb, dict):
                if "or_list" in part_sb:
                    query_or = cast(QuerySignboxOr, part_sb)
                    or_list: List[Union[str, List[str]]] = query_or["or_list"]
                    or_strs = []
                    for item in or_list:
                        if isinstance(item, str):
                            or_strs.append(item)
                        elif isinstance(item, list) and len(item) == 2:
                            or_strs.append(f"R{item[0]}{item[1]}")
                    out = "o".join(or_strs)
                elif "symbol" in part_sb:
                    query_symbol = cast(QuerySignboxSymbol, part_sb)
                    out = query_symbol["symbol"]
                elif (
                    "range" in part_sb
                    and isinstance(part_sb["range"], list)
                    and len(part_sb["range"]) == 2
                ):  # QuerySignboxRange
                    query_range = part_sb
                    out = f"R{query_range['range'][0]}{query_range['range'][1]}"

                coord = part_sb.get("coord")
                if isinstance(coord, list) and len(coord) == 2:
                    out += coord_to_swu(coord)

            query += out

    variance = swu_query_object.get("variance")
    if isinstance(variance, int):
        query += f"V{variance}"

    if swu_query_object.get("style"):
        query += "-"

    match = re.match(f"^({swuquery_pattern_full})$", query)
    return match.group(1) if match else None


# ----------------------------
# SWU to Query
# ----------------------------


def swu_to_query(swu_sign: str, flags: str) -> Optional[str]:
    """
    Function to convert an SWU sign to a query string.

    For the flags parameter, use one or more of the following:
    - A: exact symbol in temporal prefix
    - a: general symbol in temporal prefix
    - S: exact symbol in spatial signbox
    - s: general symbol in spatial signbox
    - L: spatial signbox symbol at location

    Args:
        swu_sign: SWU sign
        flags: flags for query string creation

    Returns:
        SWU query string

    Example:
        >>> swu_to_query('𝠀񀀒񀀚񋚥񋛩𝠃𝤟𝤩񋛩𝣵𝤐񀀒𝤇𝣤񋚥𝤐𝤆񀀚𝣮𝣭', 'ASL')
        'QA񀀒񀀚񋚥񋛩T񋛩𝣵𝤐񀀒𝤇𝣤񋚥𝤐𝤆񀀚𝣮𝣭'
    """
    parsed = swu_parse_sign(swu_sign)
    if not parsed.get("box"):
        return None

    a_flag = "A" in flags
    a_gen_flag = "a" in flags
    s_flag = "S" in flags
    s_gen_flag = "s" in flags
    l_flag = "L" in flags

    query = ""
    if (a_flag or a_gen_flag) and "sequence" in parsed:
        query += "A"
        for sym in parsed["sequence"]:
            query += sym + ("fr" if a_gen_flag else "")
        query += "T"

    if (s_flag or s_gen_flag) and "spatials" in parsed:
        for spatial in parsed["spatials"]:
            spatial_sym = spatial["symbol"]
            query += spatial_sym + ("fr" if s_gen_flag else "")
            if l_flag:
                query += coord_to_swu(spatial["coord"])

    return "Q" + query if query else None


# ------------------------------------------------------------------ #
# SWU Query Range
# ------------------------------------------------------------------ #


def swuquery_range(min_char: str, max_char: str) -> str:
    """
    Function to transform a range of SWU characters to a regular expression.

    Args:
        min_char: an SWU character
        max_char: an SWU character

    Returns:
        a regular expression that matches a range of SWU characters

    Example:
        >>> swuquery_range('񀀁', '񀇡')
        '[\\U00040001-\\U000401E1]'

        >>> swuquery_range('𝣔', '𝤸')
        '[\\U0001D8D4-\\U0001D938]'
    """
    if not (isinstance(min_char, str) and isinstance(max_char, str)):
        return ""
    if len(min_char) != 1 or len(max_char) != 1:
        return ""
    min_code = ord(min_char)
    max_code = ord(max_char)
    if min_code > max_code:
        return ""
    return f"[\\U{min_code:08X}-\\U{max_code:08X}]"


# ------------------------------------------------------------------ #
# SWU Query Regex
# ------------------------------------------------------------------ #


def _regex_range(sym_range: str) -> str:
    """
    Convert an SWU range (e.g., 'R񀀁񀇡') to a regex pattern.
    """
    from_key = swu_to_key(sym_range[1:2])
    to_key = swu_to_key(sym_range[-1:])
    from_char = key_to_swu(from_key[:4] + "00")
    to_char = key_to_swu(to_key[:4] + "5f")
    return swuquery_range(from_char, to_char)


def _regex_symbol(sym: str) -> str:
    """
    Convert an SWU symbol (possibly followed by 'f', 'r', or 'fr')
    into a regex covering all matching fills and/or rotations.

    Rules:
      - No suffix: match exactly this symbol
      - 'r': any rotation (same fill)
      - 'f': any fill (same rotation)
      - 'fr' or 'rf': any fill and rotation (full 96-symbol block)

    Examples:
        symbol_ranges("񀀁")   -> exact
        symbol_ranges("񀀁r")  -> contiguous 16 symbols
        symbol_ranges("񀀑f")  -> same rotation, all fills
        symbol_ranges("񀀑fr") -> all fills & rotations
    """
    base_char = sym[0]
    suffix = sym[1:] if len(sym) > 1 else ""
    base_code = ord(base_char)

    # Normalize to base symbol (fill=0, rotation=0)
    # Determine base by clearing fill/rotation offsets within the 96-symbol block.
    # Each symbol block starts every 0x60 codepoints.
    base_block = (base_code - 0x40001) // 0x60  # which symbol group
    base_start = 0x40001 + base_block * 0x60

    # Compute current offset within that block
    offset = base_code - base_start
    fill_index = offset // 0x10  # 0–5
    rot_index = offset % 0x10  # 0–15

    # total sizes
    rotation_block = 0x10
    total_block = 0x60  # 6 fills * 16 rotations = 96

    if suffix == "":
        # exact match only
        return f"\\U{base_code:08X}"

    elif suffix == "r":
        # same fill, any rotation → contiguous block of 16
        start = base_start + fill_index * rotation_block
        end = start + 0x0F
        return f"[\\U{start:08X}-\\U{end:08X}]"

    elif suffix == "f":
        # same rotation, any fill → six codepoints spaced by 0x10
        parts = [
            f"\\U{(base_start + rot_index + i * rotation_block):08X}" for i in range(6)
        ]
        return f"(?:{'|'.join(parts)})"

    elif suffix in ("fr", "rf"):
        # any fill, any rotation → contiguous 96-symbol block
        start = base_start
        end = base_start + total_block - 1
        return f"[\\U{start:08X}-\\U{end:08X}]"

    else:
        raise ValueError(f"Invalid suffix for SWU symbol: {sym}")


def swuquery_regex(query: str) -> List[str]:
    """
    Function to transform an SWU query string to one or more regular expressions.

    Args:
        query: an SWU query string

    Returns:
        a list of one or more regular expression strings

    Example:
        >>> swuquery_regex('QA񀀒T')
        ['(?:\\U0001d800\\U00040012(?:\\U00040000|(?:[\\U00040001-\\U0004f480]))*)\\U0001d80[1-4](?:\\U0001d8[0c-f][0-9a-f]|\\U0001d9[0-9a-f][0-9a-f]){2}(?:(?:[\\U00040001-\\U0004f480])(?:\\U0001d8[0c-f][0-9a-f]|\\U0001d9[0-9a-f][0-9a-f]){2})*']
    """
    pattern = re.compile(rf"^{swuquery_pattern_full}$")
    m = pattern.match(query)
    if not m:
        return []
    query = m.group(0)

    q_style = f"(?:{style_pattern_full})?"

    if query == "Q":
        return [
            f"{swu_pattern_prefix}?{swu_pattern_box}{swu_pattern_coord}(?:{swu_pattern_symbol}{swu_pattern_coord})*"
        ]
    if query == "Q-":
        return [
            f"{swu_pattern_prefix}?{swu_pattern_box}{swu_pattern_coord}(?:{swu_pattern_symbol}{swu_pattern_coord})*(?:{style_pattern_full})?"
        ]
    if query == "QT":
        return [
            f"{swu_pattern_prefix}{swu_pattern_box}{swu_pattern_coord}(?:{swu_pattern_symbol}{swu_pattern_coord})*"
        ]
    if query == "QT-":
        return [
            f"{swu_pattern_prefix}{swu_pattern_box}{swu_pattern_coord}(?:{swu_pattern_symbol}{swu_pattern_coord})*(?:{style_pattern_full})?"
        ]

    segments = []
    sortable = query.find("T") + 1
    q_sortable = ""
    if sortable:
        q_sortable = f"{swu_pattern_sort}"
        qat = query[:sortable]
        query = query[sortable:]
        if qat == "QT":
            q_sortable += f"{swu_pattern_null_or_symbol}+)"
        else:
            matches = re.findall(swuquery_pattern_list, qat)
            if matches:
                for part in matches:
                    or_list = []
                    matches_or = re.findall(swuquery_pattern_item, part)
                    if matches_or:
                        for mor in matches_or:
                            if re.match(rf"^{swuquery_pattern_range}$", mor):
                                or_list.append(_regex_range(mor))
                            else:
                                or_list.append(_regex_symbol(mor))
                    if len(or_list) == 1:
                        q_sortable += or_list[0]
                    else:
                        q_sortable += f"(?:{'|'.join(or_list)})"
                q_sortable += f"{swu_pattern_null_or_symbol}*"

    fuzz = 20
    matches_var = re.search(swuquery_pattern_var, query)
    if matches_var:
        fuzz = int(matches_var.group(0)[1:])

    item_pattern = re.compile(rf"{swuquery_pattern_symbol}|{swuquery_pattern_range}")
    matches = re.findall(rf"{swuquery_pattern_list}{swuquery_pattern_coord}", query)
    if matches:
        for item in matches:
            or_list = []
            matches_or = item_pattern.findall(item)
            for mor in matches_or:
                if re.match(rf"^{swuquery_pattern_range}$", mor):
                    or_list.append(_regex_range(mor))
                else:
                    or_list.append(_regex_symbol(mor))
            if len(or_list) == 1:
                segment = or_list[0]
            else:
                segment = f"(?:{'|'.join(or_list)})"

            coord_match = re.search(rf"{swuquery_pattern_coord}$", item)
            if coord_match and coord_match.group(0):
                coord = swu_to_coord(coord_match.group(0))
                x, y = coord
                segment += swuquery_range(num_to_swu(x - fuzz), num_to_swu(x + fuzz))
                segment += swuquery_range(num_to_swu(y - fuzz), num_to_swu(y + fuzz))
            else:
                segment += swu_pattern_coord

            segment = f"{swu_pattern_signbox}{segment}(?:{swu_pattern_symbol}{swu_pattern_coord})*"
            if sortable:
                segment = q_sortable + segment
            else:
                segment = f"{swu_pattern_prefix}?{segment}"
            if "-" in query:
                segment += q_style
            segments.append(segment)

    if not segments:
        segment = q_sortable + swu_pattern_signbox
        if "-" in query:
            segment += q_style
        segments.append(segment)

    return segments


def swuquery_results(query: str, text: str) -> List[str]:
    """
    Function that uses a query string to match signs from a string of text.

    Args:
        query: an SWU query string
        text: a string of text containing multiple signs

    Returns:
        an array of SWU signs

    Example:
        >>> swuquery_results('QA񀀒T','𝠀񀀒񀀚񋚥񋛩𝠃𝤟𝤩񋛩𝣵𝤐񀀒𝤇𝣤񋚥𝤐𝤆񀀚𝣮𝣭 𝠀񂇢񂇈񆙡񋎥񋎵𝠃𝤛𝤬񂇈𝤀𝣺񂇢𝤄𝣻񋎥𝤄𝤗񋎵𝤃𝣟񆙡𝣱𝣸 𝠀񅨑񀀙񆉁𝠃𝤙𝤞񀀙𝣷𝤀񅨑𝣼𝤀񆉁𝣳𝣮')
        ['𝠀񀀒񀀚񋚥񋛩𝠃𝤟𝤩񋛩𝣵𝤐񀀒𝤇𝣤񋚥𝤐𝤆񀀚𝣮𝣭']
    """
    if not text:
        return []

    re_patterns = swuquery_regex(query)
    if not re_patterns:
        return []

    for pattern in re_patterns:
        matches = re.findall(pattern, text)
        if matches:
            text = " ".join(matches)
        else:
            text = ""
            break  # No need to continue if no matches

    if text:
        words = list(dict.fromkeys(text.split(" ")))  # Dedup preserving order
        return [w for w in words if w]  # Filter empty
    return []


def swuquery_lines(query: str, text: str) -> List[str]:
    """
    Function that uses an SWU query string to match signs from multiple lines of text.

    Args:
        query: an SWU query string
        text: multiple lines of text, each starting with an SWU sign

    Returns:
        an array of lines of text, each starting with an SWU sign

    Example:
        >>> swuquery_lines('QA񀀒T', '''𝠀񀀒񀀚񋚥񋛩𝠃𝤟𝤩񋛩𝣵𝤐񀀒𝤇𝣤񋚥𝤐𝤆񀀚𝣮𝣭 line one
        ... 𝠀񂇢񂇈񆙡񋎥񋎵𝠃𝤛𝤬񂇈𝤀𝣺񂇢𝤄𝣻񋎥𝤄𝤗񋎵𝤃𝣟񆙡𝣱𝣸 line two
        ... 𝠀񅨑񀀙񆉁𝠃𝤙𝤞񀀙𝣷𝤀񅨑𝣼𝤀񆉁𝣳𝣮 line three''')
        ['𝠀񀀒񀀚񋚥񋛩𝠃𝤟𝤩񋛩𝣵𝤐񀀒𝤇𝣤񋚥𝤐𝤆񀀚𝣮𝣭 line one']
    """
    if not text:
        return []

    re_patterns = swuquery_regex(query)
    if not re_patterns:
        return []

    for pattern in re_patterns:
        pattern_with_anchor = f"^{pattern}.*"
        matches = re.findall(pattern_with_anchor, text, re.MULTILINE)
        if matches:
            text = "\n".join(matches)
        else:
            text = ""
            break  # No need to continue if no matches

    if text:
        lines_list = text.split("\n")
        unique_lines = list(dict.fromkeys(lines_list))  # Dedup preserving order
        return [line for line in unique_lines if line]  # Filter empty
    return []


__all__ = [
    "swuquery_parse",
    "swuquery_compose",
    "swu_to_query",
    "swuquery_range",
    "swuquery_regex",
    "swuquery_results",
    "swuquery_lines",
]



"""
The fswquery module contains functions for handling the FSW query language.

Query Language definition: https://datatracker.ietf.org/doc/id/draft-slevinski-formal-signwriting-10.html#name-query-language
"""

import re
from typing import List, Optional, Union, cast

from .datatypes import (
    QueryObject,
    QueryPrefixElement,
    QueryPrefix,
    QuerySignboxElement,
    QuerySignboxOr,
    QuerySignboxRange,
    QuerySignboxSymbol,
)

from .convert import fsw_to_coord, coord_to_fsw

from .regex import (
    style_pattern_full,
    fsw_pattern_prefix,
    fsw_pattern_signbox,
    fsw_pattern_null_or_symbol,
    fsw_pattern_symbol,
    fsw_pattern_coord,
    fswquery_pattern_var,
    fswquery_pattern_list,
    fswquery_pattern_item,
    fswquery_pattern_symbol,
    fswquery_pattern_range,
    fswquery_pattern_full,
    fswquery_pattern_coord,
    fswquery_pattern_null_or_symbol,
)

from .fsw import (
    fsw_parse_sign,
)


# ----------------------------
# FSW Query Parsing
# ----------------------------


def _fswquery_parse_prefix(text: str) -> QueryPrefix:
    if text == "T":
        return {"required": True}
    parts = re.findall(fswquery_pattern_list, text)
    processed_parts: List[QueryPrefixElement] = []
    for part in parts:
        if "o" in part:
            or_parts = re.findall(fswquery_pattern_item, part)
            processed = ["or_list"]
            for or_part in or_parts:
                if or_part.startswith("S"):
                    processed.append(or_part)
                else:
                    processed.append(or_part[1:].split("t"))
            processed_parts.append(processed)
        else:
            if part.startswith("S"):
                processed_parts.append(part)
            else:
                processed_parts.append(part[1:].split("t"))
    return {"required": True, "parts": processed_parts}


def _fswquery_parse_signbox(text: str) -> List[QuerySignboxElement]:
    """
    Parse a signbox query string into a list of query elements.

    Args:
        text: A string representing the signbox query.

    Returns:
        A list of QuerySignboxElement (symbols, ranges, or OR groups).
    """
    items = re.findall(rf"{fswquery_pattern_list}{fswquery_pattern_coord}", text)
    processed: List[QuerySignboxElement] = []

    for item in items:
        coord_str = item[-7:] if "x" in item[-7:] else ""
        front = item[:-7] if coord_str else item
        coord = fsw_to_coord(coord_str) if coord_str else None

        if "o" in front:
            or_parts = front.split("o")
            or_list: List[Union[str, List[str]]] = []
            for or_part in or_parts:
                if "S" in or_part:
                    or_list.append(or_part)
                else:
                    or_list.append(or_part[1:].split("t"))
            query_or: QuerySignboxOr = {"or_list": or_list}
            if coord is not None:
                query_or["coord"] = coord
            processed.append(query_or)
        elif "S" in front:
            query_symbol: QuerySignboxSymbol = {"symbol": front}
            if coord is not None:
                query_symbol["coord"] = coord
            processed.append(query_symbol)
        else:
            query_range: QuerySignboxRange = {"range": front[1:].split("t")}
            if coord is not None:
                query_range["coord"] = coord
            processed.append(query_range)

    return processed


def fswquery_parse(fsw_query_string: str) -> QueryObject:
    """
    Parse an FSW query string to a structured dictionary.

    Args:
        fsw_query_string: an FSW query string

    Returns:
        Dictionary representing the query structure

    Example:
        >>> fswquery_parse('QAS10000S10500oS20500oR2fft304TS100uuR205t206oS207uu510x510V5-')
        {'query': True,
         'prefix': {'required': True,
                    'parts': ['S10000',
                              ['or_list', 'S10500', 'S20500', ['2ff', '304']]]},
         'signbox': [{'symbol': 'S100uu'},
                     {'or_list': [['205', '206'], 'S207uu'], 'coord': [510, 510]}],
         'variance': 5,
         'style': True}
    """
    pattern = re.compile(rf"^{fswquery_pattern_full}$")
    m = pattern.match(fsw_query_string)
    if not m:
        return {"query": False}

    result: QueryObject = {"query": True}
    prefix = _fswquery_parse_prefix(m.group(1)) if m.group(1) else None
    if prefix is not None:
        result["prefix"] = prefix
    signbox = _fswquery_parse_signbox(m.group(2)) if m.group(2) else None
    if signbox is not None:
        result["signbox"] = signbox
    variance = int(m.group(3)[1:]) if m.group(3) else None
    if variance is not None:
        result["variance"] = variance
    style = True if m.group(4) else None
    if style is not None:
        result["style"] = style

    return result


# ----------------------------
# FSW Query Compose
# ----------------------------


def fswquery_compose(fsw_query_object: QueryObject) -> Optional[str]:
    """
    Function to compose FSW query string from object.

    Args:
        fsw_query_object: Dictionary of type QueryObject

    Returns:
        FSW query string

    Example:
        >>> fswquery_compose({
        ...     'query': True,
        ...     'prefix': {
        ...         'required': True,
        ...         'parts': [
        ...             'S10000',
        ...             ['100', '204'],
        ...             'S20500'
        ...         ]
        ...     },
        ...     'signbox': [
        ...         {'symbol': 'S20000'},
        ...         {'range': ['100', '105'], 'coord': [500, 500]}
        ...     ],
        ...     'variance': 5,
        ...     'style': True
        ... })
        'QAS10000R100t204S20500TS20000R100t105500x500V5-'
    """
    if not fsw_query_object or not fsw_query_object.get("query"):
        return None

    query = "Q"

    prefix_obj = fsw_query_object.get("prefix")
    if prefix_obj and prefix_obj.get("required"):
        parts = prefix_obj.get("parts")
        if isinstance(parts, list):
            query += "A"
            for part in parts:
                if isinstance(part, str):
                    query += part
                elif isinstance(part, list) and len(part) == 2:
                    query += f"R{part[0]}t{part[1]}"
                elif isinstance(part, list) and len(part) > 2 and part[0] == "or_list":
                    or_parts = part[1:]
                    or_strs = []
                    for or_part in or_parts:
                        if isinstance(or_part, str):
                            or_strs.append(or_part)
                        elif isinstance(or_part, list) and len(or_part) == 2:
                            or_strs.append(f"R{or_part[0]}t{or_part[1]}")
                    query += "o".join(or_strs)
        query += "T"

    signbox = fsw_query_object.get("signbox")
    if isinstance(signbox, list):
        for part_sb in signbox:
            out = ""
            if isinstance(part_sb, dict):
                if "or_list" in part_sb:
                    query_or = cast(QuerySignboxOr, part_sb)
                    or_list: List[Union[str, List[str]]] = query_or["or_list"]
                    or_strs = []
                    for item in or_list:
                        if isinstance(item, str):
                            or_strs.append(item)
                        elif isinstance(item, list) and len(item) == 2:
                            or_strs.append(f"R{item[0]}t{item[1]}")
                    out = "o".join(or_strs)
                elif "symbol" in part_sb:
                    query_symbol = cast(QuerySignboxSymbol, part_sb)
                    out = query_symbol["symbol"]
                elif (
                    "range" in part_sb
                    and isinstance(part_sb["range"], list)
                    and len(part_sb["range"]) == 2
                ):  # QuerySignboxRange
                    query_range = part_sb
                    out = f"R{query_range['range'][0]}t{query_range['range'][1]}"

                coord = part_sb.get("coord")
                if isinstance(coord, list) and len(coord) == 2:
                    out += coord_to_fsw(coord)

            query += out

    variance = fsw_query_object.get("variance")
    if isinstance(variance, int):
        query += f"V{variance}"

    if fsw_query_object.get("style"):
        query += "-"

    match = re.match(f"^({fswquery_pattern_full})$", query)
    return match.group(1) if match else None


# ----------------------------
# FSW to Query
# ----------------------------


def fsw_to_query(fsw_sign: str, flags: str) -> Optional[str]:
    """
    Function to convert an FSW sign to a query string.

    For the flags parameter, use one or more of the following:
    - A: exact symbol in temporal prefix
    - a: general symbol in temporal prefix
    - S: exact symbol in spatial signbox
    - s: general symbol in spatial signbox
    - L: spatial signbox symbol at location

    Args:
        fsw_sign: FSW sign
        flags: flags for query string creation

    Returns:
        FSW query string

    Example:
        >>> fsw_to_query('AS10011S10019S2e704S2e748M525x535S2e748483x510S10011501x466S2e704510x500S10019476x475', 'ASL')
        'QAS10011S10019S2e704S2e748TS2e748483x510S10011501x466S2e704510x500S10019476x475'
    """
    query = ""
    parsed = fsw_parse_sign(fsw_sign)

    if not parsed.get("box"):
        return None

    A_flag = "A" in flags
    a_flag = "a" in flags
    S_flag = "S" in flags
    s_flag = "s" in flags
    L_flag = "L" in flags

    if A_flag or a_flag or S_flag or s_flag:
        if (A_flag or a_flag) and (sequence := parsed.get("sequence")):
            query += "A"
            query += "".join(
                sym[:4] + ("uu" if a_flag else sym[4:6]) for sym in sequence
            )
            query += "T"

        if (S_flag or s_flag) and (spatials := parsed.get("spatials")):
            for spatial in spatials:
                sym = spatial["symbol"]
                query += sym[:4] + ("uu" if s_flag else sym[4:6])
                if L_flag:
                    query += "x".join(map(str, spatial["coord"]))

    return "Q" + query if query else None


# ----------------------------
# FSW Query Range
# ----------------------------


def _range_pattern(low: str, high: str, is_hex: bool) -> str:
    if low == high:
        return low
    if not is_hex:
        return f"[{low}-{high}]"
    hex_digits = "0123456789abcdef"
    low_index = hex_digits.index(low)
    high_index = hex_digits.index(high)
    range_str = hex_digits[low_index : high_index + 1]
    numeric_match = re.search(r"[0-9]+", range_str)
    alpha_match = re.search(r"[a-f]+", range_str)
    numeric = numeric_match.group(0) if numeric_match else ""
    alpha = alpha_match.group(0) if alpha_match else ""
    if numeric and alpha:
        return f"[{numeric[0]}-{numeric[-1]}{alpha[0]}-{alpha[-1]}]"
    elif numeric:
        return f"[{numeric[0]}-{numeric[-1]}]"
    elif alpha:
        return f"[{alpha[0]}-{alpha[-1]}]"
    else:
        return ""


def _regex_geq(p: str, q: str, is_hex: bool) -> str:
    digit_seq = "0123456789abcdef" if is_hex else "0123456789"
    last_digit = digit_seq[-1]
    p_index = digit_seq.index(p)
    if q == digit_seq[0]:
        return _range_pattern(p, last_digit, is_hex) + (
            "[0-9a-f]" if is_hex else "[0-9]"
        )
    else:
        q_range = _range_pattern(q, last_digit, is_hex)
        next_p = digit_seq[p_index + 1] if p_index + 1 < len(digit_seq) else None
        if next_p:
            next_range = _range_pattern(next_p, last_digit, is_hex)
            return p + q_range + "|" + next_range + ("[0-9a-f]" if is_hex else "[0-9]")
        else:
            return p + q_range


def _regex_leq(r: str, s: str, is_hex: bool) -> str:
    digit_seq = "0123456789abcdef" if is_hex else "0123456789"
    first_digit = digit_seq[0]
    r_index = digit_seq.index(r)
    if r_index > 0:
        prev_r = digit_seq[r_index - 1]
        prev_part = (
            _range_pattern(first_digit, prev_r, is_hex)
            + ("[0-9a-f]" if is_hex else "[0-9]")
            + "|"
        )
        s_range = _range_pattern(first_digit, s, is_hex)
        return prev_part + r + s_range
    else:
        s_range = _range_pattern(first_digit, s, is_hex)
        return r + s_range


def _regex_between_two_digits(s: str, t: str, is_hex: bool) -> str:
    s1, s2 = s[0], s[1]
    t1, t2 = t[0], t[1]
    digit_seq = "0123456789abcdef" if is_hex else "0123456789"
    s1_index = digit_seq.index(s1)
    t1_index = digit_seq.index(t1)
    if s1_index < t1_index:
        parts = [f"{s1}{_range_pattern(s2, digit_seq[-1], is_hex)}"]
        middle_digits = digit_seq[s1_index + 1 : t1_index]
        if middle_digits:
            parts.append(
                _range_pattern(middle_digits[0], middle_digits[-1], is_hex)
                + ("[0-9a-f]" if is_hex else "[0-9]")
            )
        parts.append(f"{t1}{_range_pattern(digit_seq[0], t2, is_hex)}")
        return "|".join(parts)
    else:
        return f"{s1}{_range_pattern(s2, t2, is_hex)}"


def fswquery_range(
    min_: Union[int, str], max_: Union[int, str], hex_: bool = False
) -> str:
    """
    Function to transform a range to a regular expression.

    Args:
        min_: either a decimal number or hexidecimal string
        max_: either a decimal number or hexidecimal string
        hex_: if True, the regular expression will match a hexidecimal range

    Returns:
        a regular expression that matches a range

    Example:
        >>> fswquery_range(500, 750)
        '(([56][0-9][0-9])|(7[0-4][0-9])|(750))'
        >>> fswquery_range('100', '10e', True)
        '10[0-9a-e]'
    """
    if isinstance(min_, int):
        min_str = f"{min_:03d}"
    else:
        min_str = f"{min_:0>3}"

    if isinstance(max_, int):
        max_str = f"{max_:03d}"
    else:
        max_str = str(max_)

    a, b, c = min_str[0], min_str[1], min_str[2]
    d, e, f = max_str[0], max_str[1], max_str[2]

    digit_seq = "0123456789abcdef" if hex_ else "0123456789"
    a_index = digit_seq.index(a)
    d_index = digit_seq.index(d)
    if a_index > d_index:
        raise ValueError("Start is greater than end")

    if a == d:
        between_pattern = _regex_between_two_digits(b + c, e + f, hex_)
        return (
            f"{a}(?:{between_pattern})"
            if "|" in between_pattern
            else f"{a}{between_pattern}"
        )
    else:
        parts = []
        geq_pattern = _regex_geq(b, c, hex_)
        parts.append(
            f"{a}(?:{geq_pattern})" if "|" in geq_pattern else f"{a}{geq_pattern}"
        )

        if a_index + 1 < d_index:
            middle_digits = digit_seq[a_index + 1 : d_index]
            middle_pattern = _range_pattern(middle_digits[0], middle_digits[-1], hex_)
            digit_class = "[0-9a-f]" if hex_ else "[0-9]"
            parts.append(middle_pattern + digit_class + digit_class)

        leq_pattern = _regex_leq(e, f, hex_)
        parts.append(
            f"{d}(?:{leq_pattern})" if "|" in leq_pattern else f"{d}{leq_pattern}"
        )

        return f"(?:{'|'.join(parts)})"


# ----------------------------
# FSW Query Regex
# ----------------------------


def _regex_symbol(sym: str) -> str:
    segment = sym[:4]
    fill = sym[4:5]
    if fill == "u":
        segment += "[0-5]"
    else:
        segment += fill
    rot = sym[5:6]
    if rot == "u":
        segment += "[0-9a-f]"
    else:
        segment += rot
    return segment


def _regex_range(sym_range: str) -> str:
    from_ = sym_range[1:4]
    to_ = sym_range[5:8]
    return "S" + fswquery_range(from_, to_, hex_=True) + "[0-5][0-9a-f]"


def fswquery_regex(query: str) -> List[str]:
    """
    Function to transform an FSW query string to one or more regular expressions.

    Args:
        query: an FSW query string

    Returns:
        a list of one or more regular expression strings

    Example:
        >>> fswquery_regex('QS100uuS20500480x520')
        ['(?:A(?:S[123][0-9a-f]{2}[0-5][0-9a-f])+)?[BLMR][0-9]{3}x[0-9]{3}(?:S[123][0-9a-f]{2}[0-5][0-9a-f][0-9]{3}x[0-9]{3})*S100[0-5][0-9a-f][0-9]{3}x[0-9]{3}(?:S[123][0-9a-f]{2}[0-5][0-9a-f][0-9]{3}x[0-9]{3})*',
         '(?:A(?:S[123][0-9a-f]{2}[0-5][0-9a-f])+)?[BLMR][0-9]{3}x[0-9]{3}(?:S[123][0-9a-f]{2}[0-5][0-9a-f][0-9]{3}x[0-9]{3})*S20500(?:4[6-9][0-9]|500)x(?:5[0-3][0-9]|540)(?:S[123][0-9a-f]{2}[0-5][0-9a-f][0-9]{3}x[0-9]{3})*']
    """
    pattern = re.compile(rf"^{fswquery_pattern_full}$")
    m = pattern.match(query)
    if not m:
        return []
    query = m.group(0)

    q_style = f"(?:{style_pattern_full})?"

    if query == "Q":
        return [f"{fsw_pattern_prefix}?{fsw_pattern_signbox}"]
    if query == "Q-":
        return [f"{fsw_pattern_prefix}?{fsw_pattern_signbox}{q_style}"]
    if query == "QT":
        return [f"{fsw_pattern_prefix}{fsw_pattern_signbox}"]
    if query == "QT-":
        return [f"{fsw_pattern_prefix}{fsw_pattern_signbox}{q_style}"]

    segments = []
    sortable = query.find("T") + 1
    q_sortable = ""
    if sortable:
        q_sortable = "(?:A"
        qat = query[:sortable]
        query = query[sortable:]
        if qat == "QT":
            q_sortable += f"{fsw_pattern_null_or_symbol}+)"
        else:
            matches = re.findall(fswquery_pattern_list, qat)
            if matches:
                for part in matches:
                    or_list = []
                    matches_or = re.findall(fswquery_pattern_item, part)
                    if matches_or:
                        for mor in matches_or:
                            matched = re.match(
                                rf"^{fswquery_pattern_null_or_symbol}$", mor
                            )
                            if matched:
                                or_list.append(_regex_symbol(matched.group(0)))
                            else:
                                or_list.append(_regex_range(mor))
                    if len(or_list) == 1:
                        q_sortable += or_list[0]
                    else:
                        q_sortable += f"(?:{'|'.join(or_list)})"
                q_sortable += f"{fsw_pattern_null_or_symbol}*)"

    fuzz = 20
    matches_var = re.search(fswquery_pattern_var, query)
    if matches_var:
        fuzz = int(matches_var.group(0)[1:])

    item_pattern = re.compile(rf"{fswquery_pattern_symbol}|{fswquery_pattern_range}")
    matches = re.findall(rf"{fswquery_pattern_list}{fswquery_pattern_coord}", query)
    if matches:
        for item in matches:
            or_list = []
            matches_or = item_pattern.findall(item)
            for mor in matches_or:
                if re.match(rf"^{fswquery_pattern_symbol}$", mor):
                    or_list.append(_regex_symbol(mor))
                else:
                    or_list.append(_regex_range(mor))
            if len(or_list) == 1:
                segment = or_list[0]
            else:
                segment = f"(?:{'|'.join(or_list)})"

            if "x" in item[-7:]:
                coord_str = item[-7:]
                coord = fsw_to_coord(coord_str)
                x, y = coord
                segment += fswquery_range(x - fuzz, x + fuzz)
                segment += "x"
                segment += fswquery_range(y - fuzz, y + fuzz)
            else:
                segment += fsw_pattern_coord

            segment = f"{fsw_pattern_signbox}{segment}(?:{fsw_pattern_symbol}{fsw_pattern_coord})*"
            if sortable:
                segment = q_sortable + segment
            else:
                segment = f"{fsw_pattern_prefix}?{segment}"
            if "-" in query:
                segment += q_style
            segments.append(segment)

    if not segments:
        segment = q_sortable + fsw_pattern_signbox
        if "-" in query:
            segment += q_style
        segments.append(segment)

    return segments


# ----------------------------
# FSW Query Results
# ----------------------------


def fswquery_results(query: str, text: str) -> List[str]:
    """
    Function that uses a query string to match signs from a string of text.

    Args:
        query: an FSW query string
        text: a string of text containing multiple signs

    Returns:
        an array of FSW signs

    Example:
        >>> fswquery_results('QAS10011T','AS10011S10019S2e704S2e748M525x535S2e748483x510S10011501x466S2e704510x500S10019476x475 AS15a21S15a07S21100S2df04S2df14M521x538S15a07494x488S15a21498x489S2df04498x517S2df14497x461S21100479x486 AS1f010S10018S20600M519x524S10018485x494S1f010490x494S20600481x476')
        ['AS10011S10019S2e704S2e748M525x535S2e748483x510S10011501x466S2e704510x500S10019476x475']
    """
    if not text:
        return []

    re_patterns = fswquery_regex(query)
    if not re_patterns:
        return []

    for pattern in re_patterns:
        matches = re.findall(pattern, text)
        if matches:
            text = " ".join(matches)
        else:
            text = ""
            break  # No need to continue if no matches

    if text:
        words = list(dict.fromkeys(text.split(" ")))  # Dedup preserving order
        return [w for w in words if w]  # Filter empty
    return []


def fswquery_lines(query: str, text: str) -> List[str]:
    """
    Function that uses an FSW query string to match signs from multiple lines of text.

    Args:
        query: an FSW query string
        text: multiple lines of text, each starting with an FSW sign

    Returns:
        an array of lines of text, each starting with an FSW sign

    Example:
        >>> fswquery_lines('QAS10011T',`AS10011S10019S2e704S2e748M525x535S2e748483x510S10011501x466S2e704510x500S10019476x475 line one
        ... AS15a21S15a07S21100S2df04S2df14M521x538S15a07494x488S15a21498x489S2df04498x517S2df14497x461S21100479x486 line two
        ... AS1f010S10018S20600M519x524S10018485x494S1f010490x494S20600481x476 line three`)
        ['AS10011S10019S2e704S2e748M525x535S2e748483x510S10011501x466S2e704510x500S10019476x475 line one']
    """
    if not text:
        return []

    re_patterns = fswquery_regex(query)
    if not re_patterns:
        return []

    for pattern in re_patterns:
        pattern_with_anchor = f"^{pattern}.*"
        matches = re.findall(pattern_with_anchor, text, re.MULTILINE)
        if matches:
            text = "\n".join(matches)
        else:
            text = ""
            break  # No need to continue if no matches

    if text:
        lines_list = text.split("\n")
        unique_lines = list(dict.fromkeys(lines_list))  # Dedup preserving order
        return [line for line in unique_lines if line]  # Filter empty
    return []


__all__ = [
    "fswquery_parse",
    "fswquery_compose",
    "fsw_to_query",
    "fswquery_range",
    "fswquery_regex",
    "fswquery_results",
    "fswquery_lines",
]



"""
The convert module contains functions to convert between Formal SignWriitng in ASCII (FSW) and SignWriting in Unicode (SWU) characters, along with other types of data.

Characters set definitions: https://datatracker.ietf.org/doc/id/draft-slevinski-formal-signwriting-10.html#name-character-sets
"""

from __future__ import annotations

import re
from typing import List, TypeVar, Optional, Union

from .symid_arr import symid_arr
from .regex import (
    fsw_pattern_prefix,
    fsw_pattern_box,
    fsw_pattern_coord,
    fsw_pattern_spatial,
    swu_pattern_null_or_symbol,
    swu_pattern_coord,
)


# ----------------------------
# Utility function
# ----------------------------

T = TypeVar("T")


def drop_none(d: dict[str, T | None]) -> dict[str, T]:
    """
    Remove key-value pairs from a dictionary where the value is None.

    Args:
        d: Input dictionary with optional values.

    Returns:
        Dictionary with same keys and non-None values.

    Example:
        >>> drop_none({"a": 1, "b": None, "c": "x"})
        {"a": 1, "c": "x"}
    """
    return {k: v for k, v in d.items() if v is not None}


def to_zoom(zoom: Optional[Union[int, float, str]]) -> float:
    """Convert zoom to float, handling 'x' and None."""
    if zoom == "x" or zoom is None:
        return 1.0
    return float(zoom)


# ----------------------------
# Convert functions
# ----------------------------


_SWU_TO_MARK: dict[str, str] = {
    "\U0001D800": "A",
    "\U0001D801": "B",
    "\U0001D802": "L",
    "\U0001D803": "M",
    "\U0001D804": "R",
}
_MARK_TO_SWU: dict[str, str] = {v: k for k, v in _SWU_TO_MARK.items()}


def swu_to_mark(swu_mark: str) -> str:
    """
    Convert an SWU structural marker to FSW equivalent.

    Args:
        swu_mark: single-character SWU marker

    Returns:
        single-letter FSW marker

    Example:
        >>> swu_to_mark('𝠀')
        'A'
    """
    return _SWU_TO_MARK.get(swu_mark, "")


def mark_to_swu(fsw_mark: str) -> str:
    """
    Convert an FSW structural marker to SWU equivalent.

    Args:
        fsw_mark: single-letter FSW marker

    Returns:
        single-character SWU marker

    Example:
        >>> mark_to_swu('A')
        '𝠀'
    """
    return _MARK_TO_SWU.get(fsw_mark, "")


def swu_to_num(swu_num: str) -> int:
    """
    Convert an SWU number character to an integer.

    Args:
        swu_num: a single SWU number character

    Returns:
        integer value

    Example:
        >>> swu_to_num('𝤆')
        500
    """
    return ord(swu_num) - 0x1D80C + 250


def num_to_swu(num: int) -> str:
    """
    Convert an integer to an SWU number character.

    Args:
        num: integer

    Returns:
        single SWU number character

    Example:
        >>> num_to_swu(500)
        '𝤆'
    """
    return chr(0x1D80C + int(num) - 250)


def swu_to_coord(swu_coord: str) -> List[int]:
    """
    Convert two SWU number characters to an [x, y] integer list.

    Args:
        swu_coord: string of two SWU number characters

    Returns:
        [x, y] integer list

    Example:
        >>> swu_to_coord('𝤆𝤆')
        [500, 500]
    """
    if len(swu_coord) != 2:
        return []
    return [swu_to_num(swu_coord[0]), swu_to_num(swu_coord[1])]


def coord_to_swu(coord: List[int]) -> str:
    """
    Convert an [x, y] integer list to two SWU number characters.

    Args:
        coord: [x, y] integer list

    Returns:
        string of two SWU number characters

    Example:
        >>> coord_to_swu([500, 500])
        '𝤆𝤆'
    """
    return "".join(num_to_swu(n) for n in coord)


def fsw_to_coord(fsw_coord: str) -> List[int]:
    """
    Convert an FSW coordinate string to [x, y] integer list.

    Args:
        fsw_coord: fsw coordinate string

    Returns:
        [x, y] integer list

    Example:
        >>> fsw_to_coord('500x500')
        [500, 500]
    """
    parts = fsw_coord.split("x")
    if len(parts) != 2:
        return []
    try:
        return [int(parts[0]), int(parts[1])]
    except ValueError:
        return []


def coord_to_fsw(coord: List[int]) -> str:
    """
    Convert an [x, y] list to an FSW coordinate string ('500x500').

    Args:
        coord: [x, y] integer list

    Returns:
        fsw coordinate string

    Example:
        >>> coord_to_fsw([500, 500])
        '500x500'
    """
    return "x".join(str(int(n)) for n in coord)


def swu_to_code(swu_sym: str) -> int:
    """
    Convert an SWU symbol character to its Unicode code point.

    Args:
        swu_sym: SWU symbol character

    Returns:
        integer codepoint

    Example:
        >>> swu_to_code('񀀁')
        0x40001
    """
    return ord(swu_sym)


def code_to_swu(code: int) -> str:
    """
    Convert a codepoint on plane 4 to an SWU symbol character.

    Args:
        code: integer codepoint

    Returns:
        SWU symbol character

    Example:
        >>> code_to_swu(0x40001)
        '񀀁'
    """
    return chr(int(code))


def swu_to_id(swu_sym: str) -> int:
    """
    Convert an SWU symbol character to a 16-bit ID.

    Args:
        swu_sym: SWU symbol character

    Returns:
        16-bit ID

    Example:
        >>> swu_to_id('񀀁')
        1
    """
    return swu_to_code(swu_sym) - 0x40000


def id_to_swu(id_: int) -> str:
    """
    Convert a 16-bit ID to an SWU symbol character.

    Args:
        id_: 16-bit ID

    Returns:
        SWU symbol character

    Example:
        >>> id_to_swu(1)
        '񀀁'
    """
    return code_to_swu(int(id_) + 0x40000)


def key_to_id(key: str) -> int:
    """
    Convert an FSW symbol key to a 16-bit ID.

    Args:
        key: FSW symbol key

    Returns:
        16-bit ID

    Example:
        >>> key_to_id('S10000')
        1
    """
    if key == "S00000":
        return 0
    base = int(key[1:4], 16)
    fill = int(key[4:5], 16)
    rot = int(key[5:6], 16)
    return 1 + ((base - 0x100) * 96) + (fill * 16) + rot


def id_to_key(id_: int) -> str:
    """
    Convert a 16-bit ID to an FSW symbol key.

    Args:
        id_: 16-bit ID

    Returns:
        FSW symbol key

    Example:
        >>> id_to_key(1)
        'S10000'
    """
    if int(id_) == 0:
        return "S00000"
    symcode = int(id_) - 1
    base = symcode // 96
    fill = (symcode - base * 96) // 16
    rot = symcode - base * 96 - fill * 16
    return f"S{base + 0x100:x}{fill:x}{rot:x}"


def swu_to_key(swu_sym: str) -> str:
    """
    Convert an SWU symbol character to an FSW symbol key.

    Args:
        swu_sym: SWU symbol character

    Returns:
        FSW symbol key

    Example:
        >>> swu_to_key('񀀁')
        'S10000'
    """
    if swu_sym == "\U00040000":
        return "S00000"
    symcode = swu_to_code(swu_sym) - 0x40001
    base = symcode // 96
    fill = (symcode - base * 96) // 16
    rot = symcode - base * 96 - fill * 16
    return f"S{base + 0x100:x}{fill:x}{rot:x}"


def key_to_swu(key: str) -> str:
    """
    Convert an FSW symbol key to an SWU symbol character.

    Args:
        key: FSW symbol key

    Returns:
        SWU symbol character

    Example:
        >>> key_to_swu('S10000)
        '񀀁'
    """
    if key == "S00000":
        return "\U00040000"
    base = int(key[1:4], 16)
    fill = int(key[4:5], 16)
    rot = int(key[5:6], 16)
    code = 0x40001 + ((base - 0x100) * 96) + (fill * 16) + rot
    return chr(code)


def swu_to_fsw(swu_text: str) -> str:
    """
    Convert SWU text to FSW text.

    Args:
        swu_text: SWU text string

    Returns:
        SWU text string

    Example:
        >>> swu_to_fsw('𝠀񀀒񀀚񋚥񋛩𝠃𝤟𝤩񋛩𝣵𝤐񀀒𝤇𝣤񋚥𝤐𝤆񀀚𝣮𝣭')
        'AS10011S10019S2e704S2e748M525x535S2e748483x510S10011501x466S2e704510x500S10019476x475'
    """
    if not swu_text:
        return ""

    fsw = swu_text
    for swu_c, mark in _SWU_TO_MARK.items():
        fsw = fsw.replace(swu_c, mark)

    symbols = re.findall(swu_pattern_null_or_symbol, fsw)
    for sym in symbols:
        fsw = fsw.replace(sym, swu_to_key(sym), 1)

    coords = re.findall(swu_pattern_coord, fsw)
    for coord in coords:
        fsw = fsw.replace(coord, coord_to_fsw(swu_to_coord(coord)), 1)

    return fsw


def fsw_to_swu(fsw_text: str) -> str:
    """
    Convert FSW text to SWU text.

    Args:
        fsw_text: FSW text string

    Returns:
        SWU text string

    Example:
        >>> fsw_to_swu('AS10011S10019S2e704S2e748M525x535S2e748483x510S10011501x466S2e704510x500S10019476x475')
        '𝠀񀀒񀀚񋚥񋛩𝠃𝤟𝤩񋛩𝣵𝤐񀀒𝤇𝣤񋚥𝤐𝤆񀀚𝣮𝣭'
    """
    if not fsw_text:
        return ""

    text = fsw_text

    prefixes = re.findall(fsw_pattern_prefix, text)
    for prefix in prefixes:
        keys = [prefix[i : i + 6] for i in range(1, len(prefix), 6)]
        swu_seq = "".join(key_to_swu(k) for k in keys)
        text = text.replace(prefix, "\U0001D800" + swu_seq, 1)

    box_coords = re.findall(fsw_pattern_box + fsw_pattern_coord, text)
    for box in box_coords:
        mark = box[0]
        coord_str = box[1:8]
        replacement = mark_to_swu(mark) + coord_to_swu(fsw_to_coord(coord_str))
        text = text.replace(box, replacement, 1)

    spatials = re.findall(fsw_pattern_spatial, text)
    for spatial in spatials:
        key = spatial[:6]
        coord_str = spatial[6:13]
        replacement = key_to_swu(key) + coord_to_swu(fsw_to_coord(coord_str))
        text = text.replace(spatial, replacement, 1)

    return text


def symid_max(symid_min: str) -> str:
    """
    Convert base or full symbol ID Min to symbol ID Max.

    Args:
        symid_min: minimized symbol ID

    Returns:
        maximized symbol ID

    Examples:
        >>> symid_max('101011')
        '01-01-001-01'
        >>> symid_max('101011616')
        '01-01-001-01-06-16'
    """
    if not re.fullmatch(r"\d{6}(\d{3})?", symid_min):
        return ""
    parts = [symid_min[i] for i in range(6)]
    max_str = f"0{parts[0]}-{parts[1]}{parts[2]}-0{parts[3]}{parts[4]}-0{parts[5]}"
    if len(symid_min) > 6:
        extra = [symid_min[i] for i in range(6, 9)]
        max_str += f"-0{extra[0]}-{extra[1]}{extra[2]}"
    return max_str


def symid_min(symid_max: str) -> str:
    """
    Convert base or full symbol ID Max to symbol ID Min.

    Args:
        symid_max: maximized symbol ID

    Returns:
        minimized symbol ID

    Examples:
        >>> symid_min('01-01-001-01')
        '101011'
        >>> symid_min('01-01-001-01-06-16')
        '101011616'
    """
    m = re.fullmatch(r"0(\d)-(\d{2})-0(\d{2})-0(\d)(?:-0(\d)-(\d{2}))?", symid_max)
    if not m:
        return ""
    base = m.group(1) + m.group(2) + m.group(3) + m.group(4)
    return base + (m.group(5) + m.group(6) if m.group(5) else "")


def symid_to_key(symid: str) -> str:
    """
    Convert base or full symbol ID max to FSW symbol key.

    Args:
        symid: base of full maximized symbol ID

    Returns:
        FSW symbol key

    Examples:
        >>> symid_to_key('01-01-001-01')
        'S100'
        >>> symid_to_key('01-01-001-01-06-16')
        'S1005f'
    """
    m = re.fullmatch(r"0(\d)-(\d{2})-0(\d{2})-0(\d)(?:-0(\d)-(\d{2}))?", symid)
    if not m:
        return ""
    base_min = m.group(1) + m.group(2) + m.group(3) + m.group(4)
    try:
        i = symid_arr.index(base_min)
    except ValueError:
        return ""

    if m.group(5):
        fill = int(m.group(5)) - 1
        rot = int(m.group(6)) - 1
        return f"S{256 + i:x}{fill}{rot:x}"
    else:
        return f"S{256 + i:x}"


def key_to_symid(key: str) -> str:
    """
    Convert base or full FSW symbol key to maximized symbol ID.

    Args:
        key: FSW symbol key

    Returns:
        maximized symbol ID

    Examples:
        >>> key_to_symid('S100')
        '01-01-001-01'
        >>> key_to_symid('S1005f')
        '01-01-001-01-06-16'
    """
    m = re.fullmatch(
        r"S([1-3][0-9a-f]{2})(?:([0-5])([0-9a-f]))?", key, flags=re.IGNORECASE
    )
    if not m:
        return ""
    i = int(m.group(1), 16) - 256
    if i >= len(symid_arr):
        return ""

    base = symid_max(symid_arr[i])
    if m.group(3):
        fill = 1 + int(m.group(2))
        rot = int(m.group(3), 16) + 1
        return f"{base}-0{fill}-{rot:02d}"
    return base


__all__ = [
    "drop_none",
    "to_zoom",
    "swu_to_mark",
    "mark_to_swu",
    "swu_to_num",
    "num_to_swu",
    "swu_to_coord",
    "coord_to_swu",
    "fsw_to_coord",
    "coord_to_fsw",
    "swu_to_code",
    "code_to_swu",
    "swu_to_id",
    "id_to_swu",
    "key_to_id",
    "id_to_key",
    "swu_to_key",
    "key_to_swu",
    "swu_to_fsw",
    "fsw_to_swu",
    "symid_max",
    "symid_min",
    "symid_to_key",
    "key_to_symid",
]
